/*
 * Copyright (c) 2001 Markus Friedl.  All rights reserved.
 * Copyright (c) 2007 Red Hat, Inc. All rights reserved.
 * Copyright (c) 2009 Pierre Ossman for Cendio AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "includes.h"
#ifdef HAVE_LIBNSS

#include <sys/types.h>

#include <stdarg.h>
#include <string.h>
#include <unistd.h>

#include <openssl/evp.h>

#include <nss.h>
#include <keyhi.h>
#include <pk11pub.h>
#include <cert.h>
#include <secmod.h>
#include <secerr.h>

#include "xmalloc.h"
#include "key.h"
#include "log.h"
#include "misc.h"
#include "nsskeys.h"
#include "pathnames.h"

static char *
password_cb(PK11SlotInfo *slot, PRBool retry, void *arg)
{
	char *password = arg;
	if (retry || password == NULL)
		return NULL;
	
	return PL_strdup(password);
}

int
nss_init(PK11PasswordFunc pwfn)
{
	char *dbpath;
	char buf[MAXPATHLEN];

	if (NSS_IsInitialized())
		return 0;

	if ((dbpath=getenv("NSS_DB_PATH")) == NULL) {
		struct passwd *pw;
		if ((pw = getpwuid(getuid())) == NULL ||
			pw->pw_dir == NULL) {
			return -1;
		}
		snprintf(buf, sizeof(buf), "%s/%s", pw->pw_dir,
			    _PATH_SSH_USER_DIR);
		dbpath = buf;
	}

	if (NSS_Init(dbpath) != SECSuccess) {
		debug("Failed to initialize NSS library. Attempting without DB...");
		if (NSS_NoDB_Init(NULL) != SECSuccess)
			return -1;
	}

	if (pwfn == NULL) {
		pwfn = password_cb;
	}

	PK11_SetPasswordFunc(pwfn);
	
	return 0;
}

int
nss_load_module(const char *modpath)
{
	char spec[MAXPATHLEN + 40];
	SECMODModule *module;

	debug("loading PKCS#11 module '%s'", modpath);

	snprintf(spec, sizeof(spec), "library=\"%s\" name=\"Foobar\"", modpath);
	module = SECMOD_LoadUserModule(spec, NULL, PR_FALSE);
	if (!module || !module->loaded) {
		if (module)
			SECMOD_DestroyModule(module);
		return -1;
	}

	return 0;
}

static Key *
make_key_from_privkey(SECKEYPrivateKey *privk, char *password)
{
	Key *k;
	switch (SECKEY_GetPrivateKeyType(privk)) {
		case rsaKey:
			k = key_new_nss(KEY_RSA);
			break;
		case dsaKey:
			k = key_new_nss(KEY_DSA);
			break;
		default:
			return NULL;
	}
	k->nss->pubk = SECKEY_ConvertToPublicKey(privk);
	if (k->nss->pubk != NULL) {
		k->nss->privk = SECKEY_CopyPrivateKey(privk);
	}
	if (k->nss->privk != NULL) {
		if (password != NULL) {
			k->nss->privk->wincx = xstrdup(password);
		}
		return k;
	}
	key_free(k);
	return NULL;
}

static Key **
add_key_to_list(Key *k, Key **keys, size_t *i, size_t *allocated)
{
	if (*allocated < *i + 2) {
		*allocated += 16;
		keys = xrealloc(keys, *allocated, sizeof(k));
	}
	keys[*i] = k;
	(*i)++;
	keys[*i] = NULL;
	return keys;
}

static int
nss_convert_pubkey(Key *k)
{
	u_char *n;
	unsigned int len;
	char *p;

	switch (k->type) {
		case KEY_RSA:
			n = k->nss->pubk->u.rsa.modulus.data;
			len = k->nss->pubk->u.rsa.modulus.len;

			if (BN_bin2bn(n, len, k->rsa->n) == NULL) {
				fatal("nss_convert_pubkey: BN_bin2bn failed");
			}

			n = k->nss->pubk->u.rsa.publicExponent.data;
			len = k->nss->pubk->u.rsa.publicExponent.len;

			if (BN_bin2bn(n, len, k->rsa->e) == NULL) {
				fatal("nss_convert_pubkey: BN_bin2bn failed");
			}
			break;
		case KEY_DSA:
			n = k->nss->pubk->u.dsa.params.prime.data;
			len = k->nss->pubk->u.dsa.params.prime.len;

			if (BN_bin2bn(n, len, k->dsa->p) == NULL) {
				fatal("nss_convert_pubkey: BN_bin2bn failed");
			}

			n = k->nss->pubk->u.dsa.params.subPrime.data;
			len = k->nss->pubk->u.dsa.params.subPrime.len;

			if (BN_bin2bn(n, len, k->dsa->q) == NULL) {
				fatal("nss_convert_pubkey: BN_bin2bn failed");
			}

			n = k->nss->pubk->u.dsa.params.base.data;
			len = k->nss->pubk->u.dsa.params.base.len;

			if (BN_bin2bn(n, len, k->dsa->g) == NULL) {
				fatal("nss_convert_pubkey: BN_bin2bn failed");
			}

			n = k->nss->pubk->u.dsa.publicValue.data;
			len = k->nss->pubk->u.dsa.publicValue.len;

			if (BN_bin2bn(n, len, k->dsa->pub_key) == NULL) {
				fatal("nss_convert_pubkey: BN_bin2bn failed");
			}
			break;
	}

	p = key_fingerprint(k, SSH_FP_SHA1, SSH_FP_HEX);
	debug("SHA1 fingerprint %u %s", key_size(k), p);
	xfree(p);

	return 0;
}

static int
nss_authenticate(PK11SlotInfo *slot, char *password, int pwprompts, char **output)
{
	int i, quit;

	*output = NULL;

	if (!PK11_NeedLogin(slot))
		return 0;

	for (i = 0; i < pwprompts; i++) {
		SECStatus rv;
		CK_TOKEN_INFO info;

		rv = PK11_GetTokenInfo(slot, &info);
		if (rv != SECSuccess) {
			error("Failed to get information for token %s",
				PK11_GetTokenName(slot));
			return -1;
		}

		if (info.flags & CKF_USER_PIN_LOCKED) {
			error("Passphrase for token %s is locked",
				PK11_GetTokenName(slot));
			return -1;
		}

		if (info.flags & CKF_USER_PIN_FINAL_TRY)
			debug2("Final passphrase attempt for token %s",
				PK11_GetTokenName(slot));
		else if (info.flags & CKF_USER_PIN_COUNT_LOW)
			debug2("Previous failed passphrase attempt for token %s",
				PK11_GetTokenName(slot));

		if (password != NULL)
			*output = xstrdup(password);
		else {
			char *prompt;
			if (asprintf(&prompt, "Enter passphrase for token %s: ",
				PK11_GetTokenName(slot)) < 0)
				fatal("password_cb: asprintf failed");
			*output = read_passphrase(prompt, RP_ALLOW_STDIN);
		}

		if (strcmp(*output, "") == 0) {
			debug2("no passphrase given, ignoring slot");
			quit = 1;
			goto cleanup;
		}

		quit = 0;

		rv = PK11_Authenticate(slot, PR_TRUE, *output);
		if (rv == SECSuccess)
			return 0;

		switch (PORT_GetError()) {
		case SEC_ERROR_BAD_PASSWORD:
			debug2("Incorrect passphrase, try again...");
			break;
		case SEC_ERROR_INVALID_ARGS:
		case SEC_ERROR_BAD_DATA:
			debug2("Invalid passphrase, try again...");
			break;
#if HAVE_SEC_ERROR_LOCKED_PASSWORD
		case SEC_ERROR_LOCKED_PASSWORD:
			error("Unable to authenticate, token passphrase is locked");
			quit = 1;
			break;
#endif
		default:
			error("Failure while authenticating against token");
			quit = 1;
		}

cleanup:
		memset(*output, 0, strlen(*output));
		xfree(*output);
		*output = NULL;

		/* No point in retrying the same password */
		if (password != NULL)
			break;

		if (quit)
			break;
	}

	return -1;
}

static Key **
nss_find_privkeys(const char *tokenname, const char *keyname,
    char *password, int pwprompts)
{
	Key *k = NULL;
	Key **keys = NULL;
	PK11SlotList *slots;
	PK11SlotListElement *sle;
	size_t allocated = 0;
	size_t i = 0;

	if ((slots=PK11_FindSlotsByNames(NULL, NULL, tokenname, PR_TRUE)) == NULL) {
		if (tokenname == NULL) {
			debug("No NSS token found");
		} else {
			debug("NSS token not found: %s", tokenname);
		}
		return NULL;
	}
	
	for (sle = slots->head; sle; sle = sle->next) {
		SECKEYPrivateKeyList *list;
		SECKEYPrivateKeyListNode *node;
		char *tmppass;

		if (nss_authenticate(sle->slot, password, pwprompts, &tmppass) == -1)
			break;

		debug("Looking for: %s:%s", tokenname, keyname);
		list = PK11_ListPrivKeysInSlot(sle->slot, (char *)keyname,
			tmppass);
		if (list == NULL && keyname != NULL) {
			char *fooname;
			/* NSS bug workaround */
			if (asprintf(&fooname, "%s~", keyname) < 0) {
				error("nss_find_privkey: asprintf failed");
				PK11_FreeSlotList(slots);
				return NULL;
			}
			list = PK11_ListPrivKeysInSlot(sle->slot, fooname,
			tmppass);
			free(fooname);
		}
		if (list == NULL && keyname != NULL) {
			CERTCertificate *cert;
			SECKEYPrivateKey *privk;
			cert = CERT_FindCertByNickname(CERT_GetDefaultCertDB(),
				(char *)keyname);
			if (cert == NULL)
				goto cleanup;
			privk = PK11_FindPrivateKeyFromCert(sle->slot, cert, tmppass);
			CERT_DestroyCertificate(cert);
			if (privk == NULL)
				goto cleanup;
			if ((k=make_key_from_privkey(privk, tmppass)) != NULL) {
				nss_convert_pubkey(k);
				keys = add_key_to_list(k, keys, &i, &allocated);
			}
			SECKEY_DestroyPrivateKey(privk);
		} else {
			if (list == NULL)
				goto cleanup;
			for (node=PRIVKEY_LIST_HEAD(list); !PRIVKEY_LIST_END(node, list);
				node=PRIVKEY_LIST_NEXT(node))
				if ((k=make_key_from_privkey(node->key, tmppass)) != NULL) {
					nss_convert_pubkey(k);
					keys = add_key_to_list(k, keys, &i, &allocated);
				}
			SECKEY_DestroyPrivateKeyList(list);
		}
cleanup:
		if (tmppass != NULL) {
			memset(tmppass, 0, strlen(tmppass));
			xfree(tmppass);
		}
	}
	PK11_FreeSlotList(slots);

	return keys;
}

Key **
nss_get_keys(const char *tokenname, const char *keyname,
    char *password, int pwprompts, int num_modules, const char **modules)
{
	int i;
	Key **keys;

	if (nss_init(NULL) == -1) {
		error("Failed to initialize NSS library");
		return NULL;
	}

	for (i = 0;i < num_modules;i++) {
		if (nss_load_module(modules[i]) == -1) {
			error("Failed to load PKCS#11 module '%s'", modules[i]);
			return NULL;
		}
	}

	keys = nss_find_privkeys(tokenname, keyname, password, pwprompts);
	if (keys == NULL && keyname != NULL) {
		error("Cannot find key in nss, token removed");
		return NULL;
	}
#if 0
	keys = xcalloc(3, sizeof(Key *));

	if (k->type == KEY_RSA) {
		n = key_new_nss_copy(KEY_RSA1, k);

		keys[0] = n;
		keys[1] = k;
		keys[2] = NULL;
	} else {
		keys[0] = k;
		keys[1] = NULL;
	}
#endif
	return keys;
}

char *
nss_get_key_label(Key *key)
{
	char *label, *nickname;
	
	nickname = PK11_GetPrivateKeyNickname(key->nss->privk);
	label = xstrdup(nickname);
	PORT_Free(nickname);

	return label;
}

#endif /* HAVE_LIBNSS */
